import unittest
import re
import dateutil.parser

from keys import *

class Leaf(Keyable):

    def __init__(self, required=False, es_index=None, es_analyzer=None, doc=None):
        self.required = required
        self.es_index = es_index
        self.es_analyzer = es_analyzer
        self.doc = doc

    def to_dict(self):
        retv = {
            "required":self.required,
            "doc":self.doc,
            "type":self.__class__.__name__,
            "es_type":self.ES_TYPE,
            "bq_type":self.BQ_TYPE
        }
        self.add_es_var(retv, "es_analyzer", "es_analyzer", "ES_ANALYZER")
        self.add_es_var(retv, "es_index", "es_index", "ES_INDEX")
        return retv

    def to_es(self):
        retv = {"type":self.ES_TYPE}
        self.add_es_var(retv, "index", "es_index", "ES_INDEX")
        if "index" in retv:
            assert retv["index"] in self.VALID_ES_INDEXES
        self.add_es_var(retv, "analyzer", "es_analyzer", "ES_ANALYZER")
        if "analyzer" in retv:
            assert retv["analyzer"] in self.VALID_ES_ANALYZERS
        return retv
        
    def to_bigquery(self, name):
        mode = "REQUIRED" if self.required else "NULLABLE"
        return {"name":self.key_to_bq(name), "type":self.BQ_TYPE, "mode":mode}

    def to_string(self, name):
        return "%s: %s" % (self.key_to_string(name), 
                           self.__class__.__name__.lower())
        
    def print_indent_string(self, name, indent):
        val = self.key_to_string(name)
        if indent:
            tabs = "\t" * indent
            val = tabs + val
        print val
        
    def validate(self, name, value):
        if type(value) not in self.EXPECTED_CLASS:
            raise DataValidationException("class mismatch for %s: expected %s, %s has class %s",
                                          self.key_to_string(name), self.EXPECTED_CLASS, 
                                          str(value), value.__class__.__name__)
        if hasattr(self, "_validate"):
            self._validate(self.key_to_string(name), value)
            


class AnalyzedString(Leaf):
    ES_TYPE = "string"
    BQ_TYPE = "STRING"
    ES_INDEX = "analyzed"
    EXPECTED_CLASS = [str,]
    
    INVALID = 23
    VALID = "asdf"


class String(Leaf):
    ES_TYPE = "string"
    BQ_TYPE = "STRING"
    ES_INDEX = "not_analyzed"
    EXPECTED_CLASS = [str,]

    INVALID = 23
    VALID = "asdf"


class IPv4Address(Leaf):
    ES_TYPE = "ip"
    BQ_TYPE = "STRING"
    EXPECTED_CLASS = [str,]
    IP_REGEX = re.compile('(\d{1,3}\.){3}\d{1,3}')
    
    def _is_ipv4_addr(self, ip):
        return bool(self.IP_REGEX.match(ip))
    
    def _validate(self, name, value):
        if not self._is_ipv4_addr(value):
            raise DataValidationException("%s: the value %s is not a valid IPv4 address",
                    name, value)
        
    INVALID = "my string"
    VALID = "141.212.120.0"


class Integer(Leaf):
    ES_TYPE = "integer"
    BQ_TYPE = "INTEGER"
    EXPECTED_CLASS = [int,]
    
    INVALID = 8589934592
    VALID = 234234252
    
    BITS = 32

    def _validate(self, name, value):
        max_ = 2**self.BITS - 1
        min_ = -2**self.BITS + 1
        if value > max_:
            raise DataValidationException("%s: %s is larger than max (%s)",
                    name, str(value), str(max_))
        if value < min_:
            raise DataValidationException("%s: %s is smaller than min (%s)",
                    name, str(value), str(min_))


class Byte(Integer):
    ES_TYPE = "byte"
    BITS = 8
    INVALID = 2**8+5
    VALID = 34


class Short(Integer):
    ES_TYPE = "short"
    BITS = 16
    INVALID = 2**16
    VALID = 0xFFFF


class Long(Integer):
    ES_TYPE = "long"
    BQ_TYPE = "DOUBLE"
    EXPECTED_CLASS = [int,long]
    INVALID = 2l**68
    VALID = 10l
    BITS = 64
    

class Float(Leaf):
    ES_TYPE = "float"
    BQ_TYPE = "DOUBLE"
    EXPECTED_CLASS = [float,]
    INVALID = "I'm a string!"
    VALID = 10.0
    

class Double(Float):
    ES_TYPE = "double"
    BQ_TYPE = "DOUBLE"


class Boolean(Leaf):
    ES_TYPE = "boolean"
    BQ_TYPE = "BOOLEAN"
    EXPECTED_CLASS = [bool,]
    INVALID = 0
    VALID = True


class Binary(Leaf):
    ES_TYPE = "binary"
    BQ_TYPE = "STRING"
    ES_INDEX = "no"
    EXPECTED_CLASS = [str,]
    B64_REGEX = re.compile('[A-Fa-f0-9]+')
    
    def _is_base64(self, data):
        return bool(self.B64_REGEX.match(data))
    
    def _validate(self, name, value):
        if not self._is_base64(value):
            raise DataValidationException("%s: the value %s is not valid Base64",
                                          name, value)

    VALID = "03F87824"
    INVALID = "normal"
    

class IndexedBinary(Binary):
    ES_TYPE = "binary"
    BQ_TYPE = "STRING"
    ES_INDEX = "not_analyzed"


class DateTime(Leaf):
    ES_TYPE = "datetime"
    BQ_TYPE = "TIMESTAMP"
    EXPECTED_CLASS = [str, int]
    
    VALID = "Wed Jul  8 08:52:01 EDT 2015"
    INVALID = "Wed DNE  35 08:52:01 EDT 2015"
    
    def _validate(self, name, value):
        try:
            dateutil.parser.parse(value)
        except Exception, e:
            raise DataValidationException("%s: %s is not valid timestamp",
                                          name, str(value))


VALID_LEAVES = [
    DateTime,
    AnalyzedString,
    String,
    Binary,
    IndexedBinary,
    Boolean,
    Double,
    Float,
    Long,
    Short,
    Byte,
    Integer,
    IPv4Address
]



import sys
import os.path
import json

from imp import load_source

from leaves import *
from keys import *
from compounds import *

def usage():
    print "USAGE: %s command schema [file]." % sys.argv[0].split("/")[-1]
    print "Valid commands: bigquery, elasticsearch, json, text, html, validate."
    print "schema should be defined as file.py:record"
    sys.exit(1)


def main():
    if len(sys.argv) < 3:
        usage()

    path, recname = sys.argv[2].split(":")
    module = load_source("module", path)
    record = getattr(module, recname)

    command = sys.argv[1]
    if command == "bigquery":
        print record.to_bigquery()
    elif command == "elasticsearch":
        print record.to_es(recname)
    elif command == "json":
        print record.to_json()
    elif command == "html":
        print record.to_html()
    elif command == "text":
        print record.to_html()
    elif command == "validate":
        assert os.path.exists(sys.argv[3])
        with open(sys.argv[3]) :
            for line in fd:
                record.validate(json.loads(line.strip()))
    else:
        usage()


if __name__ == "__main__":
    main()



import json
import pprint

from zschema.leaves import *
from zschema.compounds import *

class LeafUnitTests(unittest.TestCase):

    def test_valid(self):
        for leaf in VALID_LEAVES:
            leaf().validate(leaf.__name__, leaf.VALID)

    def test_invalid(self):
        for leaf in VALID_LEAVES:
            try:
                leaf().validate(leaf.__name__, leaf.INVALID)
                raise Exception("invalid value did not fail for %s",
                                leaf.__name__)
            except DataValidationException:
                continue


VALID_ELASTIC_SEARCH = {
    "host": {
        "properties": {
            "443": {
                "properties": {
                    "tls": {
                        "index": "not_analyzed", 
                        "type": "string"
                    }, 
                    "heartbleed": {
                        "properties": {
                            "heartbeat_support": {
                                "type": "boolean"
                            }, 
                            "heartbleed_vulnerable": {
                                "type": "boolean"
                            }, 
                            "timestamp": {
                                "type": "datetime"
                            }
                        }
                    }
                }
            }, 
            "ipstr": {
                "type": "ip"
            }, 
            "ip": {
                "type": "long"
            }, 
            "tags": {
                "index": "not_analyzed", 
                "type": "string"
            }
        }
    }
}

VALID_BIG_QUERY = [
    {
        "fields": [
            {
                "type": "STRING", 
                "name": "tls", 
                "mode": "NULLABLE"
            }, 
            {
                "fields": [
                    {
                        "type": "BOOLEAN", 
                        "name": "heartbeat_support", 
                        "mode": "NULLABLE"
                    }, 
                    {
                        "type": "BOOLEAN", 
                        "name": "heartbleed_vulnerable", 
                        "mode": "NULLABLE"
                    }, 
                    {
                        "type": "TIMESTAMP", 
                        "name": "timestamp", 
                        "mode": "NULLABLE"
                    }
                ], 
                "type": "RECORD", 
                "name": "heartbleed", 
                "mode": "NULLABLE"
            }
        ], 
        "type": "RECORD", 
        "name": "p443", 
        "mode": "NULLABLE"
    }, 
    {
        "type": "STRING", 
        "name": "ipstr", 
        "mode": "REQUIRED"
    }, 
    {
        "type": "STRING", 
        "name": "tags", 
        "mode": "REPEATED"
    }, 
    {
        "type": "DOUBLE", 
        "name": "ip", 
        "mode": "NULLABLE"
    }
]


class CompileAndValidationTests(unittest.TestCase):

    def setUp(self):
        heartbleed = SubRecord({
            "heartbeat_support":Boolean(),
            "heartbleed_vulnerable":Boolean(),
            "timestamp":DateTime()
        })
        
        self.host = Record({
                "ipstr":IPv4Address(required=True),
                "ip":Long(),
                Port(443):SubRecord({
                    "tls":String(),
                    "heartbleed":heartbleed
                }),
                "tags":ListOf(String())
        })

    def test_bigquery(self):
        global VALID_BIG_QUERY
        r = self.host.to_bigquery()
        self.assertEqual(r, VALID_BIG_QUERY)

    def test_elasticsearch(self):
        global VALID_ELASTIC_SEARCH
        r = self.host.to_es("host")
        self.assertEqual(json.loads(r), VALID_ELASTIC_SEARCH)

    def test_validation_known_good(self):
        test = {
            "ipstr":"141.212.120.1",
            "ip":2379511809,
            "443":{
                "tls":"test"
            }
        }
        self.host.validate(test)

    def test_validation_bad_key(self):
        test = {
            "keydne":"141.212.120.1asdf",
            "ip":2379511809,
            "443":{
                "tls":"test"
            }
        }
        try:
            self.host.validate(test)
            raise Exception("validation did not fail")
        except DataValidationException:
            pass

    def test_validation_bad_value(self):
        test = {
            "ipstr":10,
            "ip":2379511809,
            "443":{
                "tls":"test"
            }
        }
        try:
            self.host.validate(test)
            raise Exception("validation did not fail")
        except DataValidationException:
            pass

    def test_merge_no_conflict(self):
        a = SubRecord({
                "a":String(),
                "b":SubRecord({
                    "c":String()
                    })
            })
        b = SubRecord({
                "d":String(),
            })
        valid = SubRecord({
                "a":String(),
                "b":SubRecord({
                    "c":String()
                    }),
                "d":String(),

        })
        self.assertEqual(a.merge(b).to_dict(), valid.to_dict())

    def test_merge_different_types(self):
        a = SubRecord({
                "a":String(),
            })
        b = SubRecord({
                "a":SubRecord({})
            })
        try:
            a.merge(b) 
            raise Exception("validation did not fail")
        except MergeConflictException:
            pass

    def test_merge_unmergable_types(self):
        a = SubRecord({
                "a":String(),
            })
        b = SubRecord({
                "a":String(),
            })
        try:
            a.merge(b) 
            raise Exception("validation did not fail")
        except MergeConflictException:
            pass

    def test_merge_recursive(self):
        a = SubRecord({
                "m":SubRecord({
                       "a":String()
                })
            })
        b = SubRecord({
                "a":String(),
                "m":SubRecord({
                       "b":String()
                 })

            })
        c = SubRecord({
                "a":String(),
                "m":SubRecord({
                       "a":String(),
                       "b":String()
                    })
            })
        self.assertEquals(a.merge(b).to_dict(), c.to_dict())

    def test_extends(self):
        host = Record({
                 "host":IPv4Address(required=True),
                 "time":DateTime(required=True),
                 "data":SubRecord({}),
                 "error":String()
               })
        banner_grab = Record({
                        "data":SubRecord({
                                   "banner":String()
                               })
                      }, extends=host) 
        tls_banner_grab = Record({
                            "data":SubRecord({
                                       "tls":SubRecord({})
                                   })
                          }, extends=banner_grab) 
        smtp_starttls = Record({
                            "data":SubRecord({
                                       "ehlo":String()
                                   })
                        }, extends=tls_banner_grab)

        valid = Record({
                  "host":IPv4Address(required=True),
                  "time":DateTime(required=True),
                  "data":SubRecord({
                            "banner":String(),
                            "tls":SubRecord({}),
                            "ehlo":String()
                         }),
                  "error":String()
                })
        self.assertEqual(smtp_starttls.to_dict(), valid.to_dict())



class Port(object):
    def __init__(self, port):
        self.port = str(port)
        
    def to_bigquery(self):
        return "p%s" % self.port
        
    def to_es(self):
        return self.port

    to_string = to_es

    def __eq__(self, other):
        if type(other) == int:
            return int(self.port).__eq__(other)
        elif type(other) == str:
            return self.port.__eq__(other)
        else:
            return self.port.__eq__(other.port)

    def __hash__(self):
        return self.port.__hash__()


class Keyable(object):

    VALID_ES_INDEXES = [
        "analyzed", # full-text
        "not_analyzed", # searchable, not full-text,
        "no", # field is not searchable
    ]
    
    
    VALID_ES_ANALYZERS = [
        "standard",   # The standard analyzer is the default analyzer that Elasticsearch uses.
                      # It is the best general choice for analyzing text that may be in any language. 
                      # It splits the text on word boundaries, as defined by the Unicode Consortium, 
                      # and removes most punctuation.
        "simple",     # The simple analyzer splits the text on anything that isn't a letter, 
                      # and lowercases the terms. It would produce
        "whitespace", # The whitespace analyzer splits the text on whitespace. It doesn't lowercase.
    ]


    @staticmethod
    def key_to_bq(o):
        if type(o) == str:
            return o
        else:
            return o.to_bigquery()

    @staticmethod
    def key_to_es(o):
        if type(o) == str:
            return o
        else:
            return o.to_es()
            
    @staticmethod
    def key_to_string(o):
        if type(o) == str:
            return o
        else:
            return o.to_string()

    def add_es_var(self, d, name, instance, default):
        if hasattr(self, instance) and getattr(self, instance):
            d[name] = getattr(self, instance)
        elif hasattr(self, default) and getattr(self, default):
            d[name] = getattr(self, default)
        return d


class DataValidationException(TypeError):
    pass

class MergeConflictException(Exception):
    pass



from zschema.keys import Port, DataValidationException
from zschema.leaves import (
    DateTime,
    AnalyzedString,
    String,
    Binary,
    IndexedBinary,
    Boolean,
    Double,
    Float,
    Long,
    Short,
    Byte,
    Integer,
    IPv4Address
)


from zschema.compounds import ListOf, SubRecord, Record

__version__ = "0.0.5"

#__all__ = ["keys", "leaves", "compounds", "tests"]



from setuptools import setup

setup(
    name = "zschema",
    description = "A schema language for JSON documents that allows validation and compilation into various database engines",
    version = "0.0.8",
    license = "Apache License, Version 2.0",
    author = "Zakir Durumeric",
    author_email = "zakird@gmail.com",
    maintainer = "Zakir Durumeric",
    maintainer_email = "zakird@gmail.com",

    keywords = "python json schema bigquery elastic search",

    install_requires = [
        "python-dateutil"
    ],

    packages = [
        "zschema"
    ],

    entry_points={
        'console_scripts': [
            'zschema = zschema.cli:main',
        ]
    }
)

